/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2011 Christian Hemp <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __GST_I2C_H__
#define __GST_I2C_H__

#include <gst/gst.h>
#include <gst/base/gstbasetransform.h>

#include <stdio.h>
#include <stdlib.h>
G_BEGIN_DECLS

/* #defines don't like whitespacey bits */
#define GST_TYPE_I2C		(gst_i2c_get_type())
#define GST_I2C(obj)		(G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_I2C,GstI2C))
#define GST_I2CDATA(obj)	(G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_I2C,GstI2CData))
#define GST_I2C_CLASS(klass)	(G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_I2C,GstI2CClass))
#define GST_IS_I2C(obj)		(G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_I2C))
#define GST_IS_I2C_CLASS(klass)	(G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_I2C))

/**
 * GstI2CMode:
 * @GST_I2C_ACCESSMODE_SET: Set Value
 * @GST_I2C_ACCESSMODE_AND: And Value
 * @GST_I2C_ACCESSMODE_OR: Or Value
 * @GST_I2C_ACCESSMODE_XOR: Xor Value
 *
 * The different ways that i2c can behave. @I2C_ACCESSMODE_SET
 * dos not mask the value.
 */
typedef enum {
  GST_I2C_ACCESSMODE_SET,
  GST_I2C_ACCESSMODE_AND,
  GST_I2C_ACCESSMODE_OR,
  GST_I2C_ACCESSMODE_XOR,
} GstI2CAccessMode;

typedef struct _GstI2C      GstI2C;
typedef struct _GstI2CClass GstI2CClass;
typedef struct _GstI2CData  GstI2CData;

struct _GstI2CData
{
  gint addr;
  guint reg;
  GstI2CAccessMode mode;
  guint value;

  GstI2CData *next;
};

struct _GstI2C
{
  //GstElement element;
  GstBaseTransform trans;

  GstPad *sinkpad, *srcpad;

  gboolean silent;

  gchar *dev;
  gchar *file_name;
  gboolean show;

  GstI2CData *i2c_data;

  gboolean failure;

  int fd;
  FILE *file;
};

struct _GstI2CClass 
{
/*  GstI2CClass parent_class;*/
 /* GstElementClass parent_class;*/
/*  GstBaseTransformClass trans_class;*/
   GstBaseTransformClass parent_class;
};

GType gst_i2c_get_type (void);

G_END_DECLS

#endif /* __GST_I2C_H__ */
