/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2011 Christian Hemp <c.hemp@phytec.de>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
*/

/**
 * SECTION:element-i2c
 *
 * FIXME:Describe i2c here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! i2c ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <gst/base/gstbasetransform.h>
//#include <gst/controller/gstcontroller.h>

#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <linux/types.h> 
#include <linux/videodev2.h>  
#include <sys/ioctl.h>

#include "gsti2c.h"

#undef   VIDIOC_DBG_G_REGISTER
#define  VIDIOC_DBG_G_REGISTER   _IOWR ('V', 80, struct v4l2_dbg_register)
#undef   VIDIOC_DBG_S_REGISTER
#define  VIDIOC_DBG_S_REGISTER    _IOW ('V', 79, struct v4l2_dbg_register)

#define V4L2_CHIP_MATCH_I2C_ADDR   2   // Match against I2C 7-bit address

GST_DEBUG_CATEGORY_STATIC (gst_i2c_debug);
#define GST_CAT_DEFAULT gst_i2c_debug

/* Filter signals and args */
enum
{
  ARG_0,
  ARG_DEV,
  ARG_ADDR,
  ARG_REG,
  ARG_MODE,
  ARG_VALUE,
  ARG_FILE,
  ARG_SHOW,
  PROP_0,
  PROP_SILENT
};

#define GST_TYPE_I2C_ACCMODE (gst_i2c_accmode_get_type())
static GType gst_i2c_accmode_get_type (void) {
  static GType type = 0;
  static const GEnumValue data[] = {
    {GST_I2C_ACCESSMODE_SET, "Set Value", "set"},
    {GST_I2C_ACCESSMODE_AND, "And Value", "and"},
    {GST_I2C_ACCESSMODE_OR,  "Or Value",  "or"},
    {GST_I2C_ACCESSMODE_XOR, "Xor Value", "xor"},
    {0, NULL, NULL},
  };

  if (!type)
    type = g_enum_register_static ("GstI2CMode", data);

  return type;
}


/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

#define _do_init(bla) \
    GST_DEBUG_CATEGORY_INIT (gst_i2c_debug, "i2c", 0, "i2c element");

#define gst_i2c_parent_class parent_class
G_DEFINE_TYPE (GstI2C, gst_i2c, GST_TYPE_BASE_TRANSFORM);

static void gst_i2c_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);

static void gst_i2c_get_property (GObject *object, guint prop_id,
    GValue *value, GParamSpec *pspec);

static void gst_i2c_finalize (GObject *object);

static gboolean gst_i2c_transform_start  (GstBaseTransform *trans);

/*static gboolean gst_i2c_set_caps (GstPad * pad, GstCaps * caps);*/
/*static GstFlowReturn gst_i2c_chain (GstPad * pad, GstBuffer * buf);*/

static gboolean is_string_okay (gchar* str);
static gboolean is_space(gchar c);
static gchar *trim (gchar *str);
static gboolean open_device(GstI2C *i2c);
static gboolean open_file(GstI2C *i2c);
static gboolean read_file(GstI2C *i2c);
static int set_register(const GstI2C *i2c, const GstI2CData *i2c_data,
    struct v4l2_dbg_register *reg, int addr_and_reg);
static int set_mode(const GstI2C *i2c, const GstI2CData *i2c_data,
    struct v4l2_dbg_register *reg);
static void print_error(const GstI2CData *i2c_data, const int err);
static void print_info (__u32 reg_val, unsigned int reg_size);
static void cleanup(GstI2C *i2c);
/* GObject vmethod implementations */

/*static void
gst_i2c_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "I2C",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    "Christian Hemp <c.hemp@phytec.de>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}*/

/* initialize the i2c's class */
static void
gst_i2c_class_init (GstI2CClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;
  GstBaseTransformClass *trans_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_i2c_set_property;
  gobject_class->get_property = gst_i2c_get_property;
  gobject_class->finalize = gst_i2c_finalize;

  g_object_class_install_property (gobject_class, ARG_DEV,
    g_param_spec_string ("dev", "Device",
      "Device Name (Default: /dev/video0)", "/dev/video0",
      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  g_object_class_install_property (gobject_class, ARG_ADDR,
    g_param_spec_string ("addr", "Chip Address",
      "Chip Address as Hex-Value (Default: 5D)", "5D",
      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  g_object_class_install_property (gobject_class, ARG_REG,
    g_param_spec_string ("reg", "Register",
      "Register number as Hex-Value (Default: 20)", "20",
      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  g_object_class_install_property (gobject_class, ARG_MODE,
    g_param_spec_enum ("mode", "Access Mode",
    "set, and, or, xor (Default: set)", GST_TYPE_I2C_ACCMODE,
    GST_I2C_ACCESSMODE_SET,
    G_PARAM_CONSTRUCT | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  g_object_class_install_property (gobject_class, ARG_VALUE,
    g_param_spec_string ("value", "Value",
      "Hex-Value (Default: 0000)", "0",
      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  g_object_class_install_property (gobject_class, ARG_FILE,
    g_param_spec_string ("file", "File",
      "Filename (Default: none)", NULL,
      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  g_object_class_install_property (gobject_class, ARG_SHOW,
    g_param_spec_boolean ("show", "Show Accesses",
      "Show accesses of i2c (Default: true)", TRUE,
      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));

  trans_class = GST_BASE_TRANSFORM_CLASS (klass);
  trans_class->start = gst_i2c_transform_start;
  trans_class->passthrough_on_same_caps = TRUE;

  gst_element_class_set_details_simple (gstelement_class,
    "i2c plugin",
    "Generic/Filter",
    "FIXME:Generic Template Filter",
    "Christian Hemp <c.hemp@phytec.de>");

  gst_element_class_add_pad_template (gstelement_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (gstelement_class,
      gst_static_pad_template_get (&sink_factory));

/*  GST_DEBUG_CATEGORY_INIT (gst_plugin_i2c_debug,"plugin", 0, "i2c plugin");*/
}

static void
gst_i2c_init (GstI2C * filter)
{
  filter->failure = FALSE;
  filter->dev = g_malloc(sizeof(gchar) * strlen("/dev/video0"));
  if (filter->dev == NULL)
  {
    g_print("Error: memory not alloc\n");
    filter->failure = TRUE;
    return;
  }
  
  strcpy (filter->dev, "/dev/video0");
  
  filter->i2c_data = malloc(sizeof(GstI2CData));
  if (filter->i2c_data == NULL)
  {
    g_print("Error: memory not alloc\n");
    filter->failure = TRUE;
    return;
  }

  filter->i2c_data->addr = 0x5d;
  filter->i2c_data->reg = 0x20;
  filter->i2c_data->mode = GST_I2C_ACCESSMODE_SET;
  filter->i2c_data->value = 0L;
  filter->i2c_data->next = NULL;
  

  filter->file_name = NULL;
  filter->show = TRUE;
}

static gboolean
is_string_okay (gchar* str)
{
  if (strspn(str,"0123456789") == strlen(str))
    return TRUE;

  if (!strstr(str, "0x") && !strstr(str,"0X"))
    return FALSE;

  if (strspn(str,"0123456789abcdefABCDEFxX") < strlen(str))
    return FALSE;

  *str++;
  *str++;

  if (strspn(str,"0123456789abcdefABCDEF") < strlen(str))
    return FALSE;

  return TRUE;
}

static void
gst_i2c_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstI2C *filter = GST_I2C (object);

  if (filter->failure)
    return;

  switch (prop_id) {
    case ARG_DEV:
      if (filter->dev != NULL)
        filter->dev = g_realloc(filter->dev, sizeof(gchar) * strlen(g_value_get_string (value)));

      if (filter->dev == NULL)
      {
        filter->failure = TRUE;
	return;
      }
      strcpy (filter->dev, g_value_get_string (value));
      break;
    case ARG_ADDR:
      filter->i2c_data->addr = strtol (g_value_get_string (value), NULL, 0);
      break;
    case ARG_REG:
      filter->i2c_data->reg = strtol (g_value_get_string (value), NULL, 0);
      break;
    case ARG_MODE:
      filter->i2c_data->mode = g_value_get_enum (value);
      break;
    case ARG_VALUE:
      filter->i2c_data->value = strtol (g_value_get_string (value), NULL, 0);
      break;
    case ARG_FILE:
      filter->file_name = g_malloc(sizeof(gchar) * 
          strlen(g_value_get_string (value)));
      if (filter->file_name == NULL)
      {
        filter->failure = TRUE;
        return;
      }
      strcpy (filter->file_name, g_value_get_string (value));
      break;
    case ARG_SHOW:
      filter->show = g_value_get_boolean (value);
      break;
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void 
gst_i2c_get_property (GObject *object, guint prop_id, GValue *value,
    GParamSpec *pspec) {
  GstI2C *i2c;
  i2c = GST_I2C (object);
  
  if (i2c->failure)
    return;

  switch (prop_id) {
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static gboolean
is_space(gchar c)
{
  switch (c) {
    case ' ':
    case '\t':
    case '\f':
    case '\n':
    case '\r':
      return TRUE;
    default:
      return FALSE;
  }

}

static gchar
*trim (gchar *str)
{
  gchar *tmp;
  while (is_space(*str)) //trim left
        str++;

  for(tmp = str + strlen(str)-1;(tmp>=str) && is_space(*tmp); --tmp);
  tmp[1] = '\0';

  return str;
}

static gboolean
open_device(GstI2C *i2c)
{
  g_print ("i2c open %s\n", i2c->dev);

  i2c->fd = open (i2c->dev, O_RDWR);
  if (i2c->fd < 0) {
    g_print ("ERROR i2c open %s\n", i2c->dev);
    return FALSE;
  }

  return TRUE;
}

static gboolean 
open_file(GstI2C *i2c)
{

  if (i2c->file_name == NULL) 
    return TRUE;

  i2c->file = fopen(i2c->file_name, "r");
  if (i2c->file == NULL) {
    g_print ("ERROR: can't open file %s.\n", i2c->file_name);

    return FALSE;
  }
    
  return TRUE;
}

static gboolean
read_file(GstI2C *i2c)
{
  GstI2CData *i2c_data;
  GstI2CData *tmp_data;
  gboolean first = TRUE;
  gchar buffer[128];
  gchar *tmp;
  gint count;

  if (i2c->file_name == NULL)
    return TRUE;

  tmp_data = g_malloc(sizeof(GstI2CData));
  if (tmp_data == NULL)
    return FALSE;

  while (fgets(buffer, sizeof(buffer), i2c->file) != NULL) {
    strcpy(buffer,trim(buffer));

    if (strlen(buffer) <= 1 || strcspn(buffer, "#") == 0)
      continue;

    tmp = strtok(buffer, "#");
    tmp = trim(tmp);
    count = 0;

    i2c_data = g_malloc(sizeof(GstI2CData));
    if (i2c_data == NULL)
    {
      g_print("ERROR by g_malloc, exit plugin\n");
      return FALSE;
    }
    i2c_data->next = NULL;

    tmp = strtok(tmp,",");
    
    i2c_data->mode = GST_I2C_ACCESSMODE_SET;
      
    while (tmp != NULL){
      if (strcspn(tmp, "#") == 0)
        break;
      
      if (count < 3)
      {
        if(!is_string_okay(tmp))
          break;

        if (count == 0)
          i2c_data->addr = strtol(tmp,NULL,0);
        else if (count == 1)
          i2c_data->reg = strtol(tmp,NULL,0);
        else 
          i2c_data->value = strtol(tmp,NULL,0);
      }
      
      if (count == 3)
      {
        switch (*tmp & 0x5F) {
          case 'A': i2c_data->mode = GST_I2C_ACCESSMODE_AND; break;
          case 'O': i2c_data->mode = GST_I2C_ACCESSMODE_OR;  break;
          case 'X': i2c_data->mode = GST_I2C_ACCESSMODE_XOR; break;
          default:  i2c_data->mode = GST_I2C_ACCESSMODE_SET;
        }
      }
      count++;
      tmp = strtok(NULL, ",");
    }

    if (count < 3)
    {
      printf("ERROR: Failure in %s\n\t %s \n",i2c->file_name,tmp);
      i2c->failure = TRUE;
      g_free(i2c_data);
      g_free(tmp_data);
      return FALSE;
    }

    if (first)
    {
      i2c->i2c_data = i2c_data;
      first = FALSE;   
    } else {
      tmp_data = i2c->i2c_data;
      while(tmp_data->next != NULL)
        tmp_data = tmp_data->next;

      tmp_data->next = g_malloc(sizeof(GstI2CData));
      tmp_data->next = i2c_data;
    }
  }

  fclose (i2c->file);
  return TRUE;
}

static int 
set_register(const GstI2C *i2c, const GstI2CData *i2c_data, 
    struct v4l2_dbg_register *reg, int addr_and_reg)
{
  reg->match.type = V4L2_CHIP_MATCH_SUBDEV;
  
  if (addr_and_reg){
    reg->match.addr = i2c_data->addr;
    reg->reg = i2c_data->reg;
  }
 
  reg->val = 0x00;
  reg->size = 0;

  return ioctl(i2c->fd, VIDIOC_DBG_G_REGISTER, reg);
}

static int
set_mode(const GstI2C *i2c, const GstI2CData *i2c_data, 
    struct v4l2_dbg_register *reg)
{
  char line [5];
  // set/and/or/xor
  switch (i2c_data->mode) {
    case GST_I2C_ACCESSMODE_SET: 
      strcpy (line, "set"); 
      reg->val =  i2c_data->value; 
      break;
    case GST_I2C_ACCESSMODE_AND: 
      strcpy (line, "and");
      reg->val &= i2c_data->value;
      break;
    case GST_I2C_ACCESSMODE_OR:  
      strcpy (line, "or");  
      reg->val |= i2c_data->value; 
      break;
    case GST_I2C_ACCESSMODE_XOR:
      strcpy (line, "xor");
      reg->val ^= i2c_data->value;
      break;
  }
  
  if (i2c->show) {
    g_print ("i2c %s 0x%04x", line, i2c_data->value, 4);
    if (i2c_data->mode != GST_I2C_ACCESSMODE_SET)
      g_print (": 0x%04x", reg->val, 4);
    g_print ("\n");
  }

  return ioctl(i2c->fd, VIDIOC_DBG_S_REGISTER, reg);
}

static void
print_error(const GstI2CData *i2c_data, const int err)
{
  g_print ("ERROR i2c get (ioctl) addr=0x%02x reg=0x%04x: %d %s\n", 
    i2c_data->addr, i2c_data->reg, err, strerror(err));
  fflush (stdout);
}

static void
print_info (__u32 reg_val, unsigned int reg_size)
{
  g_print ("i2c get: 0x%04x (Size: %u Bytes)\n", reg_val, reg_size);
  fflush (stdout);
}

static gboolean 
gst_i2c_transform_start (GstBaseTransform *trans)
{
  GstI2C *i2c;
  GstI2CData *i2c_data;
  struct v4l2_dbg_register reg;
  int ret;
  
  i2c = GST_I2C (trans);

  if (i2c->failure || !open_device (i2c) || !open_file (i2c) || !read_file(i2c)) 
    return FALSE;

  i2c_data = i2c->i2c_data;

  while (i2c_data != NULL) {
    if (i2c -> show)
      g_print ("i2c get addr=0x%02x reg=0x%04x\n", i2c_data->addr, i2c_data->reg);

    ret = set_register (i2c, i2c_data, &reg, 1);
    if (ret) 
    {
      print_error (i2c_data, errno);
      return FALSE;
    }

    if (i2c->show)
      print_info ((__u32) reg.val, reg.size);

    ret = set_mode (i2c, i2c_data, &reg);
    if (ret)
    {
      print_error (i2c_data, errno);
      return FALSE;
    }

    if (i2c -> show)
    {
      g_print ("i2c get addr=0x%02x reg=0x%04x\n", i2c_data->addr, i2c_data->reg);

      ret = set_register (i2c, i2c_data, &reg, 0);
      if (ret)
      {
        print_error (i2c_data, errno);
        return FALSE;
      }

      print_info ((__u32) reg.val, reg.size);
    }

     i2c_data = i2c_data->next;
  }

  return TRUE;
}

static void 
gst_i2c_finalize (GObject *object) {

  GstI2C *i2c;
  GstI2CData *i2c_data;
  struct v4l2_dbg_register reg;
  int ret;

  i2c = GST_I2C (object);

  if (!i2c->show || i2c->failure)
    goto out;

  i2c_data = i2c->i2c_data;

  while (i2c_data != NULL)
  {
    g_print ("i2c get addr=0x%02x reg=0x%04x\n", i2c_data->addr, i2c_data->reg);

    ret = set_register(i2c, i2c_data, &reg, 1);
    if (ret)
      print_error(i2c_data, errno);


    print_info((__u32) reg.val, reg.size);
    i2c_data = i2c_data->next;
  }

out:
  // close i2c
  g_print ("i2c close\n");
  close (i2c->fd);
  
  cleanup(i2c);
  G_OBJECT_CLASS (parent_class)->finalize (object);
/*  GST_CALL_PARENT (G_OBJECT_CLASS, finalize, (object));*/
}

static void 
cleanup (GstI2C *i2c)
{
  GstI2CData *i2c_data =  i2c->i2c_data;
  GstI2CData *tmp_i2c_data;

  if (i2c->file_name != NULL)
    g_free(i2c->file_name);

  if (i2c->dev != NULL)
    g_free(i2c->dev);

  while (i2c_data != NULL)
  {
    tmp_i2c_data = i2c_data->next;
    g_free(i2c_data);
    i2c_data = tmp_i2c_data;
  }
}

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
#if 0
static gboolean
gst_i2c_set_caps (GstPad * pad, GstCaps * caps)
{
  GstI2C *filter;
  GstPad *otherpad;

  filter = GST_I2C (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  gst_object_unref (filter);

  return gst_pad_set_caps (otherpad, caps);
}
#endif
/* chain function
 * this function does the actual processing
 */
#if 0
static GstFlowReturn
gst_i2c_chain (GstPad * pad, GstBuffer * buf)
{
  GstI2C *filter;

  filter = GST_I2C (GST_OBJECT_PARENT (pad));

  if (filter->silent == FALSE)
    g_print ("I'm plugged, therefore I'm in.\n");

  /* just push out the incoming buffer without touching it */
  return gst_pad_push (filter->srcpad, buf);
}
#endif

/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
i2c_init (GstPlugin * i2c)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template i2c' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_i2c_debug, "i2c",
      0, "Template i2c");
 
  return gst_element_register (i2c, "i2c", GST_RANK_NONE,
      GST_TYPE_I2C);
}

/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirsti2c"
#endif

/* gstreamer looks for this structure to register i2cs
 *
 * exchange the string 'Template i2c' with your i2c description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    i2c,
    "i2c",
    i2c_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
